@@ -196,7 +196,7 @@ divesystem_idive_send (divesystem_idive_device_t *device, const unsigned char co
 	packet[0] = START;
 	packet[1] = csize;
 	memcpy(packet + 2, command, csize);
-	crc = checksum_crc_ccitt_uint16 (packet, csize + 2);
+	crc = checksum_crc16_ccitt (packet, csize + 2, 0xffff);
 	packet[csize + 2] = (crc >> 8) & 0xFF;
 	packet[csize + 3] = (crc     ) & 0xFF;
 
@@ -257,7 +257,7 @@ divesystem_idive_receive (divesystem_idive_device_t *device, unsigned char answe
 
 	// Verify the checksum.
 	unsigned short crc = array_uint16_be (packet + len + 2);
-	unsigned short ccrc = checksum_crc_ccitt_uint16 (packet, len + 2);
+	unsigned short ccrc = checksum_crc16_ccitt (packet, len + 2, 0xffff);
 	if (crc != ccrc) {
 		ERROR (abstract->context, "Unexpected packet checksum.");
 		return DC_STATUS_PROTOCOL;
