@@ -496,9 +496,17 @@ oceanic_common_device_profile (dc_device_t *abstract, dc_event_progress_t *progr
 
 		// Remove padding from the profile.
 		if (layout->highmem) {
-			unsigned char *profile = profiles + offset + layout->rb_logbook_entry_size;
-			while (rb_entry_size >= PAGESIZE && array_isequal (profile + rb_entry_size - PAGESIZE, PAGESIZE, 0xFF)) {
-				rb_entry_size -= PAGESIZE;
+			unsigned char *profile = profiles + offset;
+			// profile+12 and the bottom 4 bits of profile+13 and the top 3 bits of profile+13
+			// is the number of pages of profile data until the start of the footer
+			// (not including the prepended logbook entry).
+			unsigned int high_part = array_uint16_le (profile + 12) & 0xE000;
+			unsigned int low_part = array_uint16_le (profile + 12) & 0x0FFF;
+			unsigned int npages = ((high_part >> 1) | low_part) + 1;
+			// INFO (abstract->context, "profile npages: 0x%X (%d)", npages, npages);
+			unsigned int length = npages * PAGESIZE;
+			if (rb_entry_size > length) {
+				rb_entry_size = length;
 			}
 		}
 
