@@ -40,8 +40,6 @@
 	dc_parser_isinstance((parser), &shearwater_predator_parser_vtable) || \
 	dc_parser_isinstance((parser), &shearwater_petrel_parser_vtable))
 
-// Petrel Native Format constants
-#define PNF_BLOCKSIZE	0x20
 #define LOG_RECORD_DIVE_SAMPLE     0x01
 #define LOG_RECORD_FREEDIVE_SAMPLE 0x02
 #define LOG_RECORD_OPENING_0       0x10
@@ -61,12 +59,11 @@
 #define LOG_RECORD_CLOSING_6       0x26
 #define LOG_RECORD_CLOSING_7       0x27
 #define LOG_RECORD_FINAL           0xFF
-#define NUM_BLOCK_IDS              0x28
 
-// constant for the older Predator and Predator-like formats
 #define SZ_BLOCK   0x80
 #define SZ_SAMPLE_PREDATOR  0x10
 #define SZ_SAMPLE_PETREL    0x20
+#define SZ_SAMPLE_FREEDIVE  0x08
 
 #define GASSWITCH     0x01
 #define PPO2_EXTERNAL 0x02
@@ -79,23 +76,29 @@
 
 #define NGASMIXES 10
 #define MAXSTRINGS 32
+#define NRECORDS  7
 
 #define PREDATOR 2
 #define PETREL   3
 
+#define UNDEFINED 0xFFFFFFFF
+
 typedef struct shearwater_predator_parser_t shearwater_predator_parser_t;
 
 struct shearwater_predator_parser_t {
 	dc_parser_t base;
 	unsigned int model;
 	unsigned int petrel;
-	unsigned int pnf;
 	unsigned int samplesize;
 	// Cached fields.
 	unsigned int cached;
+	unsigned int pnf;
 	unsigned int logversion;
 	unsigned int headersize;
 	unsigned int footersize;
+	unsigned int opening[NRECORDS];
+	unsigned int closing[NRECORDS];
+	unsigned int final;
 	unsigned int ngasmixes;
 	unsigned int oxygen[NGASMIXES];
 	unsigned int helium[NGASMIXES];
@@ -103,9 +106,9 @@ struct shearwater_predator_parser_t {
 	double calibration[3];
 	unsigned int serial;
 	dc_divemode_t mode;
-
-	/* Block addresses for PNF */
-	unsigned int block_offset[NUM_BLOCK_IDS];
+	unsigned int units;
+	unsigned int atmospheric;
+	unsigned int density;
 
 	/* String fields */
 	dc_field_string_t strings[MAXSTRINGS];
@@ -116,6 +119,8 @@ static dc_status_t shearwater_predator_parser_get_datetime (dc_parser_t *abstrac
 static dc_status_t shearwater_predator_parser_get_field (dc_parser_t *abstract, dc_field_type_t type, unsigned int flags, void *value);
 static dc_status_t shearwater_predator_parser_samples_foreach (dc_parser_t *abstract, dc_sample_callback_t callback, void *userdata);
 
+static dc_status_t shearwater_predator_parser_cache (shearwater_predator_parser_t *parser);
+
 static const dc_parser_vtable_t shearwater_predator_parser_vtable = {
 	sizeof(shearwater_predator_parser_t),
 	DC_FAMILY_SHEARWATER_PREDATOR,
@@ -184,9 +189,15 @@ shearwater_common_parser_create (dc_parser_t **out, dc_context_t *context, unsig
 
 	// Set the default values.
 	parser->cached = 0;
+	parser->pnf = 0;
 	parser->logversion = 0;
 	parser->headersize = 0;
 	parser->footersize = 0;
+	for (unsigned int i = 0; i < NRECORDS; ++i) {
+		parser->opening[i] = UNDEFINED;
+		parser->closing[i] = UNDEFINED;
+	}
+	parser->final = UNDEFINED;
 	parser->ngasmixes = 0;
 	for (unsigned int i = 0; i < NGASMIXES; ++i) {
 		parser->oxygen[i] = 0;
@@ -197,6 +208,9 @@ shearwater_common_parser_create (dc_parser_t **out, dc_context_t *context, unsig
 		parser->calibration[i] = 0.0;
 	}
 	parser->mode = DC_DIVEMODE_OC;
+	parser->units = METRIC;
+	parser->density = 1025;
+	parser->atmospheric = ATM / (BAR / 1000);
 
 	*out = (dc_parser_t *) parser;
 
@@ -225,9 +239,15 @@ shearwater_predator_parser_set_data (dc_parser_t *abstract, const unsigned char
 
 	// Reset the cache.
 	parser->cached = 0;
+	parser->pnf = 0;
 	parser->logversion = 0;
 	parser->headersize = 0;
 	parser->footersize = 0;
+	for (unsigned int i = 0; i < NRECORDS; ++i) {
+		parser->opening[i] = UNDEFINED;
+		parser->closing[i] = UNDEFINED;
+	}
+	parser->final = UNDEFINED;
 	parser->ngasmixes = 0;
 	for (unsigned int i = 0; i < NGASMIXES; ++i) {
 		parser->oxygen[i] = 0;
@@ -238,6 +258,9 @@ shearwater_predator_parser_set_data (dc_parser_t *abstract, const unsigned char
 		parser->calibration[i] = 0.0;
 	}
 	parser->mode = DC_DIVEMODE_OC;
+	parser->units = METRIC;
+	parser->density = 1025;
+	parser->atmospheric = ATM / (BAR / 1000);
 
 	return DC_STATUS_SUCCESS;
 }
@@ -246,14 +269,15 @@ shearwater_predator_parser_set_data (dc_parser_t *abstract, const unsigned char
 static dc_status_t
 shearwater_predator_parser_get_datetime (dc_parser_t *abstract, dc_datetime_t *datetime)
 {
-	const unsigned char *data = abstract->data;
-	unsigned int size = abstract->size;
 	shearwater_predator_parser_t *parser = (shearwater_predator_parser_t *) abstract;
+	const unsigned char *data = abstract->data;
 
-	if (size < 2 * SZ_BLOCK)
-		return DC_STATUS_DATAFORMAT;
+	// Cache the parser data.
+	dc_status_t rc = shearwater_predator_parser_cache (parser);
+	if (rc != DC_STATUS_SUCCESS)
+		return rc;
 
-	unsigned int ticks = array_uint32_be (data + 12);
+	unsigned int ticks = array_uint32_be (data + parser->opening[0] + 12);
 
 	if (!dc_datetime_gmtime (datetime, ticks))
 		return DC_STATUS_DATAFORMAT;
@@ -361,8 +385,8 @@ add_battery_info(shearwater_predator_parser_t *parser, const char *desc, unsigne
 static void
 add_deco_model(shearwater_predator_parser_t *parser, const unsigned char *data)
 {
-	unsigned int idx_deco_model = parser->pnf ? parser->block_offset[LOG_RECORD_OPENING_2] + 18 : 67;
-	unsigned int idx_gfs = parser->pnf ? parser->block_offset[LOG_RECORD_OPENING_3] + 5 : 85;
+	unsigned int idx_deco_model = parser->pnf ? parser->opening[2] + 18 : 67;
+	unsigned int idx_gfs = parser->pnf ? parser->opening[3] + 5 : 85;
 
 	switch	(data[idx_deco_model]) {
 	case 0:
@@ -385,7 +409,7 @@ add_battery_type(shearwater_predator_parser_t *parser, const unsigned char *data
 	if (parser->logversion < 7)
 		return;
 
-	unsigned int idx_battery_type = parser->pnf ? parser->block_offset[LOG_RECORD_OPENING_4] + 9 : 120;
+	unsigned int idx_battery_type = parser->pnf ? parser->opening[4] + 9 : 120;
 	switch (data[idx_battery_type]) {
 	case 1:
 		add_string(parser, "Battery type", "1.5V Alkaline");
@@ -418,79 +442,49 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 	if (parser->cached) {
 		return DC_STATUS_SUCCESS;
 	}
+	memset(parser->strings, 0, sizeof(parser->strings));
 
-	// the log formats are very similar - but the Petrel Native Format (PNF)
-	// is organized differently. There everything is in 32 byte (PNF_BLOCKSIZE) blocks
-	// and the offsets of various fields are different. It still seems to make sense
-	// to just all parse it in one place
-
-	// header and footer are concepts of the Predator and Predator-like formats
-	unsigned int headersize = SZ_BLOCK;
-	unsigned int footersize = SZ_BLOCK;
-
-	if (size < headersize + footersize) {
+	// Verify the minimum length.
+	if (size < 2) {
 		ERROR (abstract->context, "Invalid data length.");
 		return DC_STATUS_DATAFORMAT;
 	}
 
-	// remember if this is a Petrel Native Format download
-	// if yes, we need different ways to access the various data fields
-	// for samples it's simple, they are just offset by one (so we can use pnf as offset)
-	// for header and footer data it's more complicated because of the new block structure
-	unsigned int pnf = parser->pnf = data[0] == 0x10 ? 1 : 0;
-
-	// sanity check on the log format
-	// is this a Predator-like or Petrel-native (or Teric style) log?
-	if (parser->petrel == 0 && pnf) {
-		ERROR (abstract->context, "This is a Petrel-native log, but we claim this is a Predator");
-		return DC_STATUS_DATAFORMAT;
-	}
-
-	memset (parser->block_offset, 0, NUM_BLOCK_IDS * sizeof(unsigned int));
-	if (pnf) {
-		// find the offsets of the various header and footer blocks
-		int i = 0, j = 0;
-		while (i < size) {
-			for (j = LOG_RECORD_OPENING_0; j < NUM_BLOCK_IDS; j++) {
-				if (data[i] == j)
-					parser->block_offset[j] = i;
-				if (j == LOG_RECORD_OPENING_7)
-					j = LOG_RECORD_CLOSING_0 - 1;
-			}
-			i += PNF_BLOCKSIZE;
-		}
-	}
-	// there is a small risk we are taking here... if the log were damaged and one or
-	// more of the blocks were missing, we'll default to looking into block 0 and
-	// report bogus data. This may be worth testing for?
-
-	// Log versions before 6 weren't reliably stored in the data, but
-	// 6 is also the oldest version that we assume in our code
-	unsigned int logversion = 6;
-	if (!pnf && data[127] > 6)
-		logversion = data[127];
-	if (pnf)
-		logversion = data[parser->block_offset[LOG_RECORD_OPENING_4] + 16];
-
-	INFO(abstract->context, "Shearwater log version %u\n", logversion);
-
-	memset(parser->strings, 0, sizeof(parser->strings));
-	add_string_fmt(parser, "Logversion", "%d%s", logversion, pnf ? "(PNF)" : "");
-
-	// Adjust the footersize for the final block.
-	if (parser->petrel == 1 || array_uint16_be (data + size - footersize) == 0xFFFD) {
-		footersize += SZ_BLOCK;
+	// The Petrel Native Format (PNF) is very similar to the legacy
+	// Predator and Predator-like format. The samples are simply offset
+	// by one (so we can use pnf as the offset). For the header and
+	// footer data, it's more complicated because of the new 32 byte
+	// block structure.
+	unsigned int pnf = parser->petrel ? array_uint16_be (data) != 0xFFFF : 0;
+	unsigned int headersize = 0;
+	unsigned int footersize = 0;
+	if (!pnf) {
+		// Opening and closing blocks.
+		headersize = SZ_BLOCK;
+		footersize = SZ_BLOCK;
 		if (size < headersize + footersize) {
 			ERROR (abstract->context, "Invalid data length.");
 			return DC_STATUS_DATAFORMAT;
 		}
-	}
 
-	// if this is logversion 9 or higher, make sure this isn't a freedive, as we can't parse that
-	if (logversion > 9 && pnf) {
-		if (data[parser->block_offset[LOG_RECORD_OPENING_5] + 25] == LOG_RECORD_FREEDIVE_SAMPLE) {
-			ERROR (abstract->context, "Cannot parse freedive samples");
-			return DC_STATUS_DATAFORMAT;
+		// Adjust the footersize for the final block.
+		if (parser->petrel || array_uint16_be (data + size - footersize) == 0xFFFD) {
+			footersize += SZ_BLOCK;
+			if (size < headersize + footersize) {
+				ERROR (abstract->context, "Invalid data length.");
+				return DC_STATUS_DATAFORMAT;
+			}
+
+			parser->final = size - SZ_BLOCK;
+		}
+
+		// The Predator and Predator-like format have just one large 128
+		// byte opening and closing block. To minimize the differences
+		// with the PNF format, all record offsets are assigned the same
+		// value here.
+		for (unsigned int i = 0; i < NRECORDS; ++i) {
+			parser->opening[i] = 0;
+			parser->closing[i] = size - footersize;
 		}
 	}
 
@@ -506,76 +500,97 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 	// Transmitter battery levels
 	unsigned int t1_battery = 0, t2_battery = 0;
 
-	// the indices in the sample block are offset by 1 in PNF
-	unsigned int offset = pnf ? 0 : headersize;
-	unsigned int length = pnf ? size : size - footersize;
-
-	while (offset < length) {
-		// Ignore blocks that aren't dive samples
-		if (pnf && data[offset] != LOG_RECORD_DIVE_SAMPLE) {
-			offset += parser->samplesize;
-			continue;
-		}
+	unsigned int offset = headersize;
+	unsigned int length = size - footersize;
+	while (offset + parser->samplesize <= length) {
 		// Ignore empty samples.
 		if (array_isequal (data + offset, parser->samplesize, 0x00)) {
 			offset += parser->samplesize;
 			continue;
 		}
 
-		// Status flags.
-		unsigned int status = data[offset + 11 + pnf];
-		if ((status & OC) == 0) {
-			mode = DC_DIVEMODE_CCR;
-		}
+		// Get the record type.
+		unsigned int type = pnf ? data[offset] : LOG_RECORD_DIVE_SAMPLE;
 
-		// Gaschange.
-		unsigned int o2 = data[offset + 7 + pnf];
-		unsigned int he = data[offset + 8 + pnf];
-		if (o2 != o2_previous || he != he_previous) {
-			// Find the gasmix in the list.
-			unsigned int idx = 0;
-			while (idx < ngasmixes) {
-				if (o2 == oxygen[idx] && he == helium[idx])
-					break;
-				idx++;
+		if (type == LOG_RECORD_DIVE_SAMPLE) {
+			// Status flags.
+			unsigned int status = data[offset + 11 + pnf];
+			if ((status & OC) == 0) {
+				mode = DC_DIVEMODE_CCR;
 			}
 
-			// Add it to list if not found.
-			if (idx >= ngasmixes) {
-				if (idx >= NGASMIXES) {
-					ERROR (abstract->context, "Maximum number of gas mixes reached.");
-					return DC_STATUS_NOMEMORY;
+			// Gaschange.
+			unsigned int o2 = data[offset + 7 + pnf];
+			unsigned int he = data[offset + 8 + pnf];
+			if (o2 != o2_previous || he != he_previous) {
+				// Find the gasmix in the list.
+				unsigned int idx = 0;
+				while (idx < ngasmixes) {
+					if (o2 == oxygen[idx] && he == helium[idx])
+						break;
+					idx++;
 				}
-				oxygen[idx] = o2;
-				helium[idx] = he;
-				ngasmixes = idx + 1;
+
+				// Add it to list if not found.
+				if (idx >= ngasmixes) {
+					if (idx >= NGASMIXES) {
+						ERROR (abstract->context, "Maximum number of gas mixes reached.");
+						return DC_STATUS_NOMEMORY;
+					}
+					oxygen[idx] = o2;
+					helium[idx] = he;
+					ngasmixes = idx + 1;
+				}
+
+				o2_previous = o2;
+				he_previous = he;
 			}
 
-			o2_previous = o2;
-			he_previous = he;
-		}
-
-		// Transmitter battery levels
-		if (logversion >= 7) {
+			// Transmitter battery levels
 			// T1 at offset 27, T2 at offset 19
 			t1_battery |= battery_state(data + offset + 27 + pnf);
 			t2_battery |= battery_state(data + offset + 19 + pnf);
+		} else if (type == LOG_RECORD_FREEDIVE_SAMPLE) {
+			// Freedive record
+			mode = DC_DIVEMODE_FREEDIVE;
+		} else if (type >= LOG_RECORD_OPENING_0 && type <= LOG_RECORD_OPENING_7) {
+			// Opening record
+			parser->opening[type - LOG_RECORD_OPENING_0] = offset;
+		} else if (type >= LOG_RECORD_CLOSING_0 && type <= LOG_RECORD_CLOSING_7) {
+			// Closing record
+			parser->closing[type - LOG_RECORD_CLOSING_0] = offset;
+		} else if (type == LOG_RECORD_FINAL) {
+			// Final record
+			parser->final = offset;
 		}
 
 		offset += parser->samplesize;
 	}
 
-	// for header and footer indices we use a variable base that is set to the
-	// correct value based on the log type
-	unsigned int base = 0;
+	// Verify the required opening/closing records.
+	for (unsigned int i = 0; i < NRECORDS - 2; ++i) {
+		if (parser->opening[i] == UNDEFINED || parser->closing[i] == UNDEFINED) {
+			ERROR (abstract->context, "Opening or closing record %u not found.", i);
+			return DC_STATUS_DATAFORMAT;
+		}
+	}
+
+	// Log versions before 6 weren't reliably stored in the data, but
+	// 6 is also the oldest version that we assume in our code
+	unsigned int logversion = data[parser->opening[4] + (pnf ? 16 : 127)];
+	add_string_fmt(parser, "Logversion", "%d%s", logversion, pnf ? "(PNF)" : "");
+
+	// The transmitter battery levels are only valid for logversion 7+
+	if (logversion < 7) {
+		t1_battery = 0;
+		t2_battery = 0;
+	}
 
 	// Cache sensor calibration for later use
 	unsigned int nsensors = 0, ndefaults = 0;
-
-	// calibration value for sensors
-	base = pnf ? parser->block_offset[LOG_RECORD_OPENING_3] + 7 : 87;
+	unsigned int base = parser->opening[3] + (pnf ? 6 : 86);
 	for (size_t i = 0; i < 3; ++i) {
-		unsigned int calibration = array_uint16_be(data + base + i * 2);
+		unsigned int calibration = array_uint16_be(data + base + 1 + i * 2);
 		parser->calibration[i] = calibration / 100000.0;
 		if (parser->model == PREDATOR) {
 			// The Predator expects the mV output of the cells to be
@@ -584,7 +599,7 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 			// sensors lines up and matches the average.
 			parser->calibration[i] *= 2.2;
 		}
-		if (data[base - 1] & (1 << i)) {
+		if (data[base] & (1 << i)) {
 			if (calibration == 2100) {
 				ndefaults++;
 			}
@@ -602,12 +617,13 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 		if (mode != DC_DIVEMODE_OC)
 			add_string(parser, "PPO2 source", "voted/averaged");
 	} else {
-		parser->calibrated = data[base - 1];
+		parser->calibrated = data[base];
 		if (mode != DC_DIVEMODE_OC)
 			add_string(parser, "PPO2 source", "cells");
 	}
 
 	// Cache the data for later use.
+	parser->pnf = pnf;
 	parser->logversion = logversion;
 	parser->headersize = headersize;
 	parser->footersize = footersize;
@@ -617,6 +633,11 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 		parser->helium[i] = helium[i];
 	}
 	parser->mode = mode;
+	parser->units = data[parser->opening[0] + 8];
+	parser->atmospheric = array_uint16_be (data + parser->opening[1] + (parser->pnf ? 16 : 47));
+	parser->density = array_uint16_be (data + parser->opening[3] + (parser->pnf ? 3 : 83));
+	parser->cached = 1;
+
 	add_string_fmt(parser, "Serial", "%08x", parser->serial);
 	// bytes 1-31 are identical in all formats
 	add_string_fmt(parser, "FW Version", "%2x", data[19]);
@@ -625,7 +646,6 @@ shearwater_predator_parser_cache (shearwater_predator_parser_t *parser)
 	add_string_fmt(parser, "Battery at end", "%.1f V", data[9] / 10.0);
 	add_battery_info(parser, "T1 battery", t1_battery);
 	add_battery_info(parser, "T2 battery", t2_battery);
-	parser->cached = 1;
 
 	return DC_STATUS_SUCCESS;
 }
@@ -636,42 +656,29 @@ shearwater_predator_parser_get_field (dc_parser_t *abstract, dc_field_type_t typ
 	shearwater_predator_parser_t *parser = (shearwater_predator_parser_t *) abstract;
 
 	const unsigned char *data = abstract->data;
-	unsigned int size = abstract->size;
 
 	// Cache the parser data.
 	dc_status_t rc = shearwater_predator_parser_cache (parser);
 	if (rc != DC_STATUS_SUCCESS)
 		return rc;
 
-	// Get the offset to the footer record.
-	unsigned int footer = size - parser->footersize;
-
-	// Get the unit system.
-	unsigned int units = data[8];
-
 	dc_gasmix_t *gasmix = (dc_gasmix_t *) value;
 	dc_salinity_t *water = (dc_salinity_t *) value;
 	dc_field_string_t *string = (dc_field_string_t *) value;
-	unsigned int density = 0;
 
-	// the first 32 bytes of the footer and closing block 0 are identical
-	unsigned int block_start = parser->pnf ? parser->block_offset[LOG_RECORD_CLOSING_0] : footer;
 	if (value) {
-		unsigned int idx;
 		switch (type) {
 		case DC_FIELD_DIVETIME:
-			// FIXME: this is wrong based on the documentation I received
-			//        it should be a 3 byte value in offsets 6-8 that is dive length in seconds
-			*((unsigned int *) value) = array_uint16_be (data + block_start + 6) * 60;
+			if (parser->pnf)
+				*((unsigned int *) value) = array_uint24_be (data + parser->closing[0] + 6);
+			else
+				*((unsigned int *) value) = array_uint16_be (data + parser->closing[0] + 6) * 60;
 			break;
 		case DC_FIELD_MAXDEPTH:
-			if (units == IMPERIAL)
-				*((double *) value) = array_uint16_be (data + block_start + 4) * FEET;
+			if (parser->units == IMPERIAL)
+				*((double *) value) = array_uint16_be (data + parser->closing[0] + 4) * FEET;
 			else
-				*((double *) value) = array_uint16_be (data + block_start + 4);
-			// according to the documentation this should have been in tenth of a meter
-			// before, but the existing code for the Predator-like format didn't have
-			// that adjustment, so let's just do that for PNF (where we definitely need it).
+				*((double *) value) = array_uint16_be (data + parser->closing[0] + 4);
 			if (parser->pnf)
 				*((double *)value) /= 10.0;
 			break;
@@ -684,17 +691,14 @@ shearwater_predator_parser_get_field (dc_parser_t *abstract, dc_field_type_t typ
 			gasmix->nitrogen = 1.0 - gasmix->oxygen - gasmix->helium;
 			break;
 		case DC_FIELD_SALINITY:
-			idx = parser->pnf ? parser->block_offset[LOG_RECORD_OPENING_3] + 3 : 83;
-			density = array_uint16_be (data + idx);
-			if (density == 1000)
+			if (parser->density == 1000)
 				water->type = DC_WATER_FRESH;
 			else
 				water->type = DC_WATER_SALT;
-			water->density = density;
+			water->density = parser->density;
 			break;
 		case DC_FIELD_ATMOSPHERIC:
-			idx = parser->pnf ? parser->block_offset[LOG_RECORD_OPENING_1] + 16 : 47;
-			*((double *) value) = array_uint16_be (data + idx) / 1000.0;
+			*((double *) value) = parser->atmospheric / 1000.0;
 			break;
 		case DC_FIELD_DIVEMODE:
 			*((dc_divemode_t *) value) = parser->mode;
@@ -730,23 +734,25 @@ shearwater_predator_parser_samples_foreach (dc_parser_t *abstract, dc_sample_cal
 	if (rc != DC_STATUS_SUCCESS)
 		return rc;
 
-	// Get the unit system.
-	unsigned int units = data[8];
-
 	// Previous gas mix.
 	unsigned int o2_previous = 0, he_previous = 0;
 
+	// Sample interval.
 	unsigned int time = 0;
+	unsigned int interval = 10;
+	if (parser->pnf && parser->logversion >= 9) {
+		interval = array_uint16_be (data + parser->opening[5] + 23);
+		if (interval % 1000 != 0) {
+			ERROR (abstract->context, "Unsupported sample interval (%u ms).", interval);
+			return DC_STATUS_DATAFORMAT;
+		}
+		interval /= 1000;
+	}
+
 	unsigned int pnf = parser->pnf;
-	unsigned int offset = pnf ? 0 : parser->headersize;
-	unsigned int length = pnf ? size : size - parser->footersize;
-	unsigned int time_increment = 10;
-
-	// the time increment is now given in ms. not sure how we'll deal with that since all we do is full seconds
-	if (pnf && parser->logversion >= 9)
-		time_increment = array_uint16_be (data + parser->block_offset[LOG_RECORD_OPENING_5] + 23) / 1000;
-
-	while (offset < length) {
+	unsigned int offset = parser->headersize;
+	unsigned int length = size - parser->footersize;
+	while (offset + parser->samplesize <= length) {
 		dc_sample_value_t sample = {0};
 
 		// stop parsing if we see the end block
@@ -764,44 +770,48 @@ shearwater_predator_parser_samples_foreach (dc_parser_t *abstract, dc_sample_cal
 			continue;
 		}
 
-		// Time (seconds).
-		time += time_increment;
-		sample.time = time;
-		if (callback) callback (DC_SAMPLE_TIME, sample, userdata);
+		// Get the record type.
+		unsigned int type = pnf ? data[offset] : LOG_RECORD_DIVE_SAMPLE;
 
-		// Depth (1/10 m or ft).
-		unsigned int depth = array_uint16_be (data + pnf + offset);
-		if (units == IMPERIAL)
-			sample.depth = depth * FEET / 10.0;
-		else
-			sample.depth = depth / 10.0;
-		if (callback) callback (DC_SAMPLE_DEPTH, sample, userdata);
+		if (type == LOG_RECORD_DIVE_SAMPLE) {
+			// Time (seconds).
+			time += interval;
+			sample.time = time;
+			if (callback) callback (DC_SAMPLE_TIME, sample, userdata);
 
-		// Temperature (°C or °F).
-		int temperature = (signed char) data[offset + pnf + 13];
-		if (temperature < 0) {
-			// Fix negative temperatures.
-			temperature += 102;
-			if (temperature > 0) {
-				temperature = 0;
+			// Depth (1/10 m or ft).
+			unsigned int depth = array_uint16_be (data + pnf + offset);
+			if (parser->units == IMPERIAL)
+				sample.depth = depth * FEET / 10.0;
+			else
+				sample.depth = depth / 10.0;
+			if (callback) callback (DC_SAMPLE_DEPTH, sample, userdata);
+
+			// Temperature (°C or °F).
+			int temperature = (signed char) data[offset + pnf + 13];
+			if (temperature < 0) {
+				// Fix negative temperatures.
+				temperature += 102;
+				if (temperature > 0) {
+					temperature = 0;
+				}
 			}
-		}
-		if (units == IMPERIAL)
-			sample.temperature = (temperature - 32.0) * (5.0 / 9.0);
-		else
-			sample.temperature = temperature;
-		if (callback) callback (DC_SAMPLE_TEMPERATURE, sample, userdata);
+			if (parser->units == IMPERIAL)
+				sample.temperature = (temperature - 32.0) * (5.0 / 9.0);
+			else
+				sample.temperature = temperature;
+			if (callback) callback (DC_SAMPLE_TEMPERATURE, sample, userdata);
 
-		// Status flags.
-		unsigned int status = data[offset + pnf + 11];
+			// Status flags.
+			unsigned int status = data[offset + pnf + 11];
 
-		if ((status & OC) == 0) {
-			// PPO2
-			if ((status & PPO2_EXTERNAL) == 0) {
-				if (!parser->calibrated) {
+			if ((status & OC) == 0) {
+				// PPO2
+				if ((status & PPO2_EXTERNAL) == 0) {
+#ifdef SENSOR_AVERAGE
 					sample.ppo2 = data[offset + pnf + 6] / 100.0;
 					if (callback) callback (DC_SAMPLE_PPO2, sample, userdata);
-				} else {
+#else
 					sample.ppo2 = data[offset + pnf + 12] * parser->calibration[0];
 					if (callback && (parser->calibrated & 0x01)) callback (DC_SAMPLE_PPO2, sample, userdata);
 
@@ -810,97 +820,125 @@ shearwater_predator_parser_samples_foreach (dc_parser_t *abstract, dc_sample_cal
 
 					sample.ppo2 = data[offset + pnf + 15] * parser->calibration[2];
 					if (callback && (parser->calibrated & 0x04)) callback (DC_SAMPLE_PPO2, sample, userdata);
+#endif
 				}
-			}
 
-			// Setpoint
-			if (parser->petrel) {
-				sample.setpoint = data[offset + pnf + 18] / 100.0;
-			} else {
-				// this will only ever be called for the actual Predator, so no adjustment needed for PNF
-				if (status & SETPOINT_HIGH) {
-					sample.setpoint = data[18] / 100.0;
+				// Setpoint
+				if (parser->petrel) {
+					sample.setpoint = data[offset + pnf + 18] / 100.0;
 				} else {
-					sample.setpoint = data[17] / 100.0;
+					// this will only ever be called for the actual Predator, so no adjustment needed for PNF
+					if (status & SETPOINT_HIGH) {
+						sample.setpoint = data[18] / 100.0;
+					} else {
+						sample.setpoint = data[17] / 100.0;
+					}
+				}
+				if (callback) callback (DC_SAMPLE_SETPOINT, sample, userdata);
+			}
+
+			// CNS
+			if (parser->petrel) {
+				sample.cns = data[offset + pnf + 22] / 100.0;
+				if (callback) callback (DC_SAMPLE_CNS, sample, userdata);
+			}
+
+			// Gaschange.
+			unsigned int o2 = data[offset + pnf + 7];
+			unsigned int he = data[offset + pnf + 8];
+			if (o2 != o2_previous || he != he_previous) {
+				unsigned int idx = shearwater_predator_find_gasmix (parser, o2, he);
+				if (idx >= parser->ngasmixes) {
+					ERROR (abstract->context, "Invalid gas mix.");
+					return DC_STATUS_DATAFORMAT;
+				}
+
+				sample.gasmix = idx;
+				if (callback) callback (DC_SAMPLE_GASMIX, sample, userdata);
+				o2_previous = o2;
+				he_previous = he;
+			}
+
+			// Deco stop / NDL.
+			unsigned int decostop = array_uint16_be (data + offset + pnf + 2);
+			if (decostop) {
+				sample.deco.type = DC_DECO_DECOSTOP;
+				if (parser->units == IMPERIAL)
+					sample.deco.depth = decostop * FEET;
+				else
+					sample.deco.depth = decostop;
+			} else {
+				sample.deco.type = DC_DECO_NDL;
+				sample.deco.depth = 0.0;
+			}
+			sample.deco.time = data[offset + pnf + 9] * 60;
+			if (callback) callback (DC_SAMPLE_DECO, sample, userdata);
+
+			// for logversion 7 and newer (introduced for Perdix AI)
+			// detect tank pressure
+			if (parser->logversion >= 7) {
+				// Tank pressure
+				// Values above 0xFFF0 are special codes:
+				//    0xFFFF AI is off
+				//    0xFFFE No comms for 90 seconds+
+				//    0xFFFD No comms for 30 seconds
+				//    0xFFFC Transmitter not paired
+				// For regular values, the top 4 bits contain the battery
+				// level (0=normal, 1=critical, 2=warning), and the lower 12
+				// bits the tank pressure in units of 2 psi.
+				unsigned int pressure = array_uint16_be (data + offset + pnf + 27);
+				if (pressure < 0xFFF0) {
+					pressure &= 0x0FFF;
+					sample.pressure.tank = 0;
+					sample.pressure.value = pressure * 2 * PSI / BAR;
+					if (callback) callback (DC_SAMPLE_PRESSURE, sample, userdata);
+				}
+				pressure = array_uint16_be (data + offset + pnf + 19);
+				if (pressure < 0xFFF0) {
+					pressure &= 0x0FFF;
+					sample.pressure.tank = 1;
+					sample.pressure.value = pressure * 2 * PSI / BAR;
+					if (callback) callback (DC_SAMPLE_PRESSURE, sample, userdata);
+				}
+
+				// Gas time remaining in minutes
+				// Values above 0xF0 are special codes:
+				//    0xFF Not paired
+				//    0xFE No communication
+				//    0xFD Not available in current mode
+				//    0xFC Not available because of DECO
+				//    0xFB Tank size or max pressure haven’t been set up
+				if (data[offset + pnf + 21] < 0xF0) {
+					sample.rbt = data[offset + pnf + 21];
+					if (callback) callback (DC_SAMPLE_RBT, sample, userdata);
 				}
 			}
-			if (callback) callback (DC_SAMPLE_SETPOINT, sample, userdata);
-		}
+		} else if (type == LOG_RECORD_FREEDIVE_SAMPLE) {
+			// A freedive record is actually 4 samples, each 8-bytes,
+			// packed into a standard 32-byte sized record. At the end
+			// of a dive, unused partial records will be 0 padded.
+			for (unsigned int i = 0; i < 4; ++i) {
+				unsigned int idx = offset + i * SZ_SAMPLE_FREEDIVE;
 
-		// CNS
-		if (parser->petrel) {
-			sample.cns = data[offset + pnf + 22] / 100.0;
-			if (callback) callback (DC_SAMPLE_CNS, sample, userdata);
-		}
+				// Ignore empty samples.
+				if (array_isequal (data + idx, SZ_SAMPLE_FREEDIVE, 0x00)) {
+					break;
+				}
 
-		// Gaschange.
-		unsigned int o2 = data[offset + pnf + 7];
-		unsigned int he = data[offset + pnf + 8];
-		if (o2 != o2_previous || he != he_previous) {
-			unsigned int idx = shearwater_predator_find_gasmix (parser, o2, he);
-			if (idx >= parser->ngasmixes) {
-				ERROR (abstract->context, "Invalid gas mix.");
-				return DC_STATUS_DATAFORMAT;
-			}
+				// Time (seconds).
+				time += interval;
+				sample.time = time;
+				if (callback) callback (DC_SAMPLE_TIME, sample, userdata);
 
-			sample.gasmix = idx;
-			if (callback) callback (DC_SAMPLE_GASMIX, sample, userdata);
-			o2_previous = o2;
-			he_previous = he;
-		}
+				// Depth (absolute pressure in millibar)
+				unsigned int depth = array_uint16_be (data + idx + 1);
+				sample.depth = (depth - parser->atmospheric) * (BAR / 1000.0) / (parser->density * GRAVITY);
+				if (callback) callback (DC_SAMPLE_DEPTH, sample, userdata);
 
-		// Deco stop / NDL.
-		unsigned int decostop = array_uint16_be (data + offset + pnf + 2);
-		if (decostop) {
-			sample.deco.type = DC_DECO_DECOSTOP;
-			if (units == IMPERIAL)
-				sample.deco.depth = decostop * FEET;
-			else
-				sample.deco.depth = decostop;
-		} else {
-			sample.deco.type = DC_DECO_NDL;
-			sample.deco.depth = 0.0;
-		}
-		sample.deco.time = data[offset + pnf + 9] * 60;
-		if (callback) callback (DC_SAMPLE_DECO, sample, userdata);
-
-		// for logversion 7 and newer (introduced for Perdix AI)
-		// detect tank pressure
-		if (parser->logversion >= 7) {
-			// Tank pressure
-			// Values above 0xFFF0 are special codes:
-			//    0xFFFF AI is off
-			//    0xFFFE No comms for 90 seconds+
-			//    0xFFFD No comms for 30 seconds
-			//    0xFFFC Transmitter not paired
-			// For regular values, the top 4 bits contain the battery
-			// level (0=normal, 1=critical, 2=warning), and the lower 12
-			// bits the tank pressure in units of 2 psi.
-			unsigned int pressure = array_uint16_be (data + offset + pnf + 27);
-			if (pressure < 0xFFF0) {
-				pressure &= 0x0FFF;
-				sample.pressure.tank = 0;
-				sample.pressure.value = pressure * 2 * PSI / BAR;
-				if (callback) callback (DC_SAMPLE_PRESSURE, sample, userdata);
-			}
-			pressure = array_uint16_be (data + offset + pnf + 19);
-			if (pressure < 0xFFF0) {
-				pressure &= 0x0FFF;
-				sample.pressure.tank = 1;
-				sample.pressure.value = pressure * 2 * PSI / BAR;
-				if (callback) callback (DC_SAMPLE_PRESSURE, sample, userdata);
-			}
-
-			// Gas time remaining in minutes
-			// Values above 0xF0 are special codes:
-			//    0xFF Not paired
-			//    0xFE No communication
-			//    0xFD Not available in current mode
-			//    0xFC Not available because of DECO
-			//    0xFB Tank size or max pressure haven’t been set up
-			if (data[offset + pnf + 21] < 0xF0) {
-				sample.rbt = data[offset + pnf + 21];
-				if (callback) callback (DC_SAMPLE_RBT, sample, userdata);
+				// Temperature (1/10 °C).
+				int temperature = (signed short) array_uint16_be (data + idx + 3);
+				sample.temperature = temperature / 10.0;
+				if (callback) callback (DC_SAMPLE_TEMPERATURE, sample, userdata);
 			}
 		}
 
